
#include <strings.h>

#include <sys/file.h>
#include <sys/stropts.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/ioctl.h>

#include <net/if.h>
#include <net/nit_if.h>
#include <net/nit_pf.h>
#include <net/packetfilt.h>

#include <netinet/in.h>
#include <netinet/if_ether.h>


typedef unsigned char u_int8_t;
typedef unsigned short u_int16_t;

#include "rmp.h"
#undef RMP_MAX_PACKET
#define RMP_MAX_PACKET 1514
#include "rmp_var.h"

static struct ether_header ether_header[0];
static struct rmp_packet rmp_packet[0];

#define OFFSET(str,mem) ((char*)&(str)[0].mem-(char*)str)
#define NIT_OFFSET(str,mem) (OFFSET(str,mem)/sizeof(short))

static const struct ether_addr rmp_addr = {RMP_ADDR};


#define DEV_NIT "/dev/nit"

#define MAXNUMIF 8


char *
ether_interfaces(void)
{
  char cbuf[MAXNUMIF * sizeof (struct ifreq)];
  struct ifconf ifc;
  int sfd;
  int i, j;
  int numinter;
  static char result[IFNAMSIZ];
  static char *const validname[] = {"le", "ie", "ec", 0};


  if (result[0] != '\0') {
    return result;
  }

  if ((sfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    return 0;
  }
  ifc.ifc_len = sizeof cbuf;
  ifc.ifc_buf = cbuf;
  if (ioctl(sfd, SIOCGIFCONF, (char *)&ifc) < 0) {
    close(sfd);
    return 0;
  }
  close(sfd);

  numinter = ifc.ifc_len / sizeof (struct ifreq);
  for (i = 0; i < numinter; ++i) {
    for (j = 0; validname[j] != 0; ++j) {
      if (strncmp(ifc.ifc_req[i].ifr_name, validname[j], 2) == 0) {
	strcpy(result, ifc.ifc_req[i].ifr_name);
	return result;
      }
    }
  }
  return 0;
}



static int
i_str(int fd, int cmd, char *ptr, int len)
{
  struct strioctl si;

  si.ic_cmd = cmd;
  si.ic_timout = INFTIM;
  si.ic_len = len;
  si.ic_dp = ptr;
  return ioctl(fd, I_STR, (char *)&si);
}


int
ether_open(char *name)
{
  int fd;
  struct ifreq ifr;
  struct packetfilt filter;
  unsigned short *fptr;


  if ((fd = open (DEV_NIT, O_RDWR)) < 0) {
    return -1;
  }

  if (ioctl(fd, I_SRDOPT, (char *) RMSGD) < 0) {
    return -1;
  }

  strncpy(ifr.ifr_name, name, sizeof ifr.ifr_name);
  if (i_str(fd, NIOCBIND, (char *)&ifr, sizeof ifr) < 0) {
    return -1;
  }

  if (ioctl(fd, I_PUSH, "pf") < 0) {
    return -1;
  }

  fptr = &filter.Pf_Filter[0];

  *fptr++ = ENF_PUSHWORD + NIT_OFFSET(ether_header, ether_type);
  *fptr++ = ENF_PUSHLIT | ENF_LE;
  *fptr++ = htons(1500);
  *fptr++ = ENF_PUSHZERO | ENF_CNOR;

  *fptr++ = ENF_PUSHWORD + NIT_OFFSET(rmp_packet, hp_llc.dsap);
  *fptr++ = ENF_PUSHLIT | ENF_CAND;
  *fptr++ = htons((IEEE_DSAP_HP<<8)|IEEE_SSAP_HP);

  *fptr++ = ENF_PUSHWORD + NIT_OFFSET(rmp_packet, hp_llc.cntrl);
  *fptr++ = ENF_PUSHLIT | ENF_CAND;
  *fptr++ = htons(IEEE_CNTL_HP);

  *fptr++ = ENF_PUSHWORD + NIT_OFFSET(rmp_packet, hp_llc.filler);
  *fptr++ = ENF_PUSHZERO | ENF_CAND;

  *fptr++ = ENF_PUSHWORD + NIT_OFFSET(rmp_packet, hp_llc.dxsap);
  *fptr++ = ENF_PUSHLIT | ENF_CAND;
  *fptr++ = htons(HPEXT_DXSAP);

  *fptr++ = ENF_PUSHWORD + NIT_OFFSET(rmp_packet, hp_llc.sxsap);
  *fptr++ = ENF_PUSHLIT | ENF_CAND;
  *fptr++ = htons(HPEXT_SXSAP);

  filter.Pf_FilterLen = fptr - &filter.Pf_Filter[0];
  if (i_str(fd, NIOCSETF, (char *)&filter, sizeof (struct packetfilt)) < 0) {
    return -1;
  }

  ifr.ifr_addr.sa_family = AF_NIT;
  *(struct ether_addr *)&ifr.ifr_addr.sa_data = rmp_addr;
  if (ioctl(fd, SIOCADDMULTI, &ifr) < 0) {
    return -1;
  }

  ioctl(fd, I_FLUSH, (char *)FLUSHR);

  return fd;
}


int
ether_write(int fd, char *pkt, int len)
{
  struct sockaddr sa;
  struct strbuf cbuf;
  struct strbuf dbuf;

  sa.sa_family = AF_UNSPEC;
  *(struct ether_header *)sa.sa_data = *(struct ether_header *)pkt;

  cbuf.maxlen = cbuf.len = sizeof sa;
  cbuf.buf = (char *)&sa;

  dbuf.maxlen = dbuf.len = len;
  dbuf.buf = pkt + sizeof (struct ether_header);

  return putmsg(fd, &cbuf, &dbuf, 0);
}


int
ether_read(int fd, char *pkt, int len)
{
  return read(fd, pkt, sizeof (struct ether_header) + len);
}
